/*!
 * Kajabity Wiki Text Plugin for jQuery  
 * Version: 0.2
 * http://www.kajabity.com/jquery-wikitext/
 * 
 * Copyright 2011, Simon J. Williams
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * Kajabity is a trademark of Williams Technologies Limited.
 * 
 * http://www.williams-technologies.co.uk
 */
(function($)
{
	/**
	 * jQuery Utility Function to convert Wiki formatted 'plaintext' to HTML.
	 * 
	 * @param plaintext
	 *            the Wiki text to be converted to HTML.
	 * @return HTML formatted text.
	 */
	$.wikiText = function(plaintext)
	{
		// The source text with nulls/undefined taken care of.
		var source = (plaintext || '').toString();
		// The resultant HTML string - initially empty.
		var html = '';

		// Regular expressions to read the source one line at a time.
		var regex = /([^\r\n]*)(\r\n?|\n)/g;
		var lineMatches;
		var offset = 0;
		var line;

		// Regular expressions to match each kind of line level format mark-up.
		var re_blank = /^([ \t]*)$/;
		var re_heading = /^(={1,6})[ \t]+([^=]+)(={1,6})[ \t]*$/;
		var re_bullet = /^[ \t]+\*[ \t]+(.+)$/;
		var re_numbered = /^[ \t]+#[ \t]+(.+)$/;
		var re_mono_start = /^\{{3}$/;
		var re_mono_end = /^\}{3}$/;
		var re_blockquote = /^[ \t]+(.+)$/;
		var matches;

		// Flags indicating which kind of block we are currently in.
		var paragraph = false;
		var olist = false;
		var ulist = false;
		var bq = false;
		var mono = false;

		// Regular expression to find mark-up tokens in each line.
		var regexToken = /(((ftp|https?):\/\/)[\-\w@:%_\+.~#?,&\/\/=]+)|((mailto:)?[_.\w-]+@([\w][\w\-]+\.)+[a-zA-Z]{2,3})|((\r\n?|\n))|(!?((_{2})|(\{{3})|(\}{3})|(~{2})|(\^)|(,{2})|(\[)|('{2,5})))/g;

		// Individual mark-up regular expressions.
		var re_link = /((ftp|https?):\/\/)[\-\w@:%_\+.~#?,&\/\/=]+/;
		var re_mail = /(mailto:)?([_.\w\-]+@([\w][\w\-]+\.)+[a-zA-Z]{2,3})/;
		var tn_monospace = "{{{";
		var tn_monospace_end = "}}}";
		var tn_bolditalic = "'''''";
		var tn_bold = "'''";
		var tn_italic = "''";
		var tn_superscript = "^";
		var tn_subscript = ",,";
		var tn_underline = "__";
		var tn_strikethrough = "~~";
		var monospace = false;

		// Keep track of inline format nesting.
		var tagStack = [];
		var poppedStack = [];

		// Inline formatting start tags.
		var beginings =
		{
			bold : "<strong>",
			italic : "<em>",
			monospace : "<tt>",
			underline : "<u>",
			strikethrough : "<strike>",
			superscript : "<sup>",
			subscript : "<sub>"
		};

		// ...and end tags.
		var endings =
		{
			bold : "</strong>",
			italic : "</em>",
			monospace : "</tt>",
			underline : "</u>",
			strikethrough : "</strike>",
			superscript : "</sup>",
			subscript : "</sub>"
		};

		/**
		 * Remove inline formatting at end of a block. Puts it on the
		 * poppedStack to add at the start of the next block.
		 * 
		 * @return end tags for any current inline formatting.
		 */
		var endFormatting = function()
		{
			var tags = '';
			var popped;
			while( tagStack.length > 0 )
			{
				popped = tagStack.pop();
				tags += endings[popped];
				poppedStack.push( popped );
			}
			return tags;
		};

		/**
		 * End the current block and, temporarily, any nested inline formatting,
		 * if any.
		 * 
		 * @return the block (and inline formatting) end tags.
		 */
		var endBlock = function()
		{
			html += endFormatting();

			if( paragraph )
			{
				html += "</p>\n";
				paragraph = false;
			}
			if( olist )
			{
				html += "</li>\n</ol>\n";
				olist = false;
			}
			if( ulist )
			{
				html += "</li>\n</ul>\n";
				ulist = false;
			}
			if( bq )
			{
				html += "</blockquote>\n";
				bq = false;
			}
		};

		/**
		 * Re-add nested formatting removed at the end of the previous block.
		 * 
		 * @return start tags for all continued formatting.
		 */
		var restartFormatting = function()
		{
			var tags = '';
			while( poppedStack.length > 0 )
			{
				var popped = poppedStack.pop();
				tags += beginings[popped];
				tagStack.push( popped );
			}
			return tags;
		};

		/**
		 * As most inline format tags are the same at the start or end, this
		 * toggles the formatting on or off depending if it is currently in the
		 * tagStack.
		 * 
		 * @param label
		 *            the name of the format to toggle.
		 * @return any start or end tags to toggle the formatting with proper
		 *         nesting.
		 */
		var toggleFormatting = function(label)
		{
			var tags = '';
			if( $.inArray( label, tagStack ) > -1 )
			{
				var popped;
				do
				{
					popped = tagStack.pop();
					tags += endings[popped];
					if( popped === label )
					{
						break;
					}
					poppedStack.push( popped );
				}
				while( popped !== label );

				tags += restartFormatting();
			}
			else
			{
				tagStack.push( label );
				tags = beginings[label];
			}

			return tags;
		};

		/**
		 * Apply inline formatting to text in a line - and escape any HTML
		 * mark-up tags.
		 * 
		 * @param text
		 *            the plain text to be formatted and escaped.
		 * @return HTML formatted text.
		 */
		var formatText = function(text)
		{
			var sourceToken = (text || '').toString();
			var formattedText = '';
			var token;
			var offset = 0;
			var tokenArray;
			var linkText;

			// Iterate through any mark-up tokens in the line.
			while( (tokenArray = regexToken.exec( sourceToken )) !== null )
			{
				token = tokenArray[0];
				if( offset < tokenArray.index )
				{
					// Add non-mark-up text.
					formattedText += $.wikiText.safeText( sourceToken
							.substring( offset, tokenArray.index ) );
				}

				if( monospace )
				{
					// Ignore mark-up until end of monospace.
					if( tn_monospace_end === token )
					{
						monospace = false;
						formattedText += toggleFormatting( "monospace" );
					}
					else
					{
						formattedText += $.wikiText.safeText( token );
					}
				}
				else if( tn_monospace === token )
				{
					monospace = true;
					formattedText += toggleFormatting( "monospace" );
				}
				else if( re_link.test( token ) )
				{
					linkText = $.wikiText.safeText( decodeURI( token ) );
					formattedText += linkText.link( encodeURI( token ) )
							.toString();
				}
				else if( re_mail.test( token ) )
				{
					linkText = decodeURI( token.replace( /mailto:/, "" ) );
					formattedText += linkText.link( encodeURI( "mailto:"
							+ linkText ) );
				}
				else if( tn_bold === token )
				{
					formattedText += toggleFormatting( "bold" );
				}
				else if( tn_italic === token )
				{
					formattedText += toggleFormatting( "italic" );
				}
				else if( tn_bolditalic === token )
				{
					// Avoid empty tag if nesting is wrong way around.
					if( $.inArray( "bold", tagStack ) > $.inArray( "italic",
							tagStack ) )
					{
						formattedText += toggleFormatting( "bold" );
						formattedText += toggleFormatting( "italic" );
					}
					else
					{
						formattedText += toggleFormatting( "italic" );
						formattedText += toggleFormatting( "bold" );
					}
				}
				else if( tn_superscript === token )
				{
					formattedText += toggleFormatting( "superscript" );
				}
				else if( tn_subscript === token )
				{
					formattedText += toggleFormatting( "subscript" );
				}
				else if( tn_underline === token )
				{
					formattedText += toggleFormatting( "underline" );
				}
				else if( tn_strikethrough === token )
				{
					formattedText += toggleFormatting( "strikethrough" );
				}
				else if( token[0] === "!" )
				{
					formattedText += $.wikiText.safeText( token.substring( 1 ) );
				}
				else
				{
					formattedText += $.wikiText.safeText( token );
				}

				offset = regexToken.lastIndex;
			}

			if( offset < sourceToken.length )
			{
				// Add trailing non-mark-up text.
				formattedText += $.wikiText.safeText( sourceToken
						.substring( offset ) );
			}

			return formattedText;
		};

		/**
		 * Get a single line from the input. This resolves the issue where the
		 * last line is not returned because it doesn't end with CR/LF.
		 * 
		 * @return a single line of input - or null at end of string.
		 */
		var getLine = function()
		{
			if( offset < source.length )
			{
				lineMatches = regex.exec( source );
				if( lineMatches != null )
				{
					offset = regex.lastIndex;
					line = lineMatches[1];
				}
				else
				{
					line = source.substring( offset );
					offset = source.length;
				}
			}
			else
			{
				line = null;
			}

			return line;
		};

		// --------------------------------------------------------------------

		while( getLine() != null )
		{
			if( mono )
			{
				if( line.match( re_mono_end ) )
				{
					mono = false;
					html += "</pre>\n";
				}
				else
				{
					html += $.wikiText.safeText( line ) + "\n";
				}
			}
			else if( line.length === 0 || re_blank.test( line ) )
			{
				endBlock();
			}
			else if( (matches = line.match( re_heading )) !== null )
			{
				endBlock();
				var headingLevel = matches[1].length;
				html += "\n<h" + headingLevel + ">" + 
						restartFormatting() + 
						formatText( matches[2] ) + 
						endFormatting() + 
						"</h" + headingLevel + ">\n\n";
			}
			else if( (matches = line.match( re_bullet )) !== null )
			{
				if( ulist )
				{
					html += endFormatting() + "</li>\n";
				}
				else
				{
					endBlock();
					html += "<ul>\n";
					ulist = true;
				}

				html += "<li>" + restartFormatting() + formatText( matches[1] );
			}
			else if( (matches = line.match( re_numbered )) !== null )
			{
				if( olist )
				{
					html += endFormatting() + "</li>\n";
				}
				else
				{
					endBlock();
					html += "<ol>\n";
					olist = true;
				}

				html += "<li>" + restartFormatting() + formatText( matches[1] );
			}
			else if( line.match( re_mono_start ) )
			{
				endBlock();
				html += "<pre>\n";
				mono = true;
			}
			else if( (matches = line.match( re_blockquote )) )
			{
				// If not already in blockquote - or a list...
				if( !(bq || olist || ulist) )
				{
					endBlock();
					html += "<blockquote>\n";
					html += restartFormatting();
					bq = true;
				}

				html += "\n" + formatText( matches[1] );
			}
			else
			{
				if( !paragraph )
				{
					endBlock();
					html += "<p>\n";
					html += restartFormatting();
					paragraph = true;
				}

				html += formatText( line ) + "\n";
			}
		}

		endBlock();

		return html;
	};

	/**
	 * Escape HTML special characters.
	 * 
	 * @param text
	 *            which may contain HTML mark-up characters.
	 * @return text with HTML mark-up characters escaped.
	 */
	$.wikiText.safeText = function(text)
	{
		return text.replace( /&/g, "&amp;" ).replace( /</g, "&lt;" ).replace(
				/>/g, "&gt;" );
	};

	/**
	 * A jQuery Wrapper Function to append Wiki formatted text to a DOM object
	 * converted to HTML.
	 * 
	 * @param plaintext
	 *            text with Wiki mark-up.
	 * @return the wrapped set.
	 */
	$.fn.wikiText = function(plaintext)
	{
		return this.html( $.wikiText( plaintext ) );
	};
})( jQuery );
